# coding: utf-8

__name__        = "kvaser"
__version__     = "0.1.5"
__license__     = "Apache Software License"
__description__ = "Various utility functions."
__author__      = u"Klaus Kähler Holst"
__email__       = "klaus@holst.it"
__url__         = "https://github.com/kkholst/kvaser"
__copyright__   = u"Copyright 2019-2022, Klaus Kähler Holst"
__summary__     = "Utility functions"
__keywords__    = "tools, utilities"

__all__ = [
    "__name__", "__summary__", "__description__",
    "__url__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]



# -*- coding: utf-8 -*-
#
# Data examples
# Copyright (c) 2019-2022 Klaus K. Holst.  All rights reserved.

import pkg_resources
import pandas as pd


def getdata(dataset:str='d', list:bool=False):
    """Load example data from the 'kvaser' package

    Parameters
    ----------
    filename: str
       Name of the filename of the data
    list: bool

    Examples
    --------
    >>> # Surgical unit data
    >>> kvaser.getdata('surgunit').head()
       bloodclot  prognostic  enzyme  liver function  survival
    0        6.7          62      81            2.59       200
    1        5.1          59      66            1.70       101
    2        7.4          57      83            2.16       204
    3        6.5          73      41            2.01       101
    4        7.8          65     115            4.30       509
    >>> # Example data for risk-regression model
    >>> kvaser.getdata().head()
       y  a         x         z
    0  0  0 -0.626454  1.134965
    1  0  0  0.183643  1.111932
    2  0  0 -0.835629 -0.870778
    3  1  0  1.595281  0.210732
    4  1  1  0.329508  0.069396

    """
    if list:
        return(pkg_resources.resource_listdir('kvaser', 'data'))
    filename = 'data/' + dataset + '.csv.gz'
    inp = pkg_resources.resource_filename('kvaser', filename)
    data = pd.read_csv(inp, sep=',', header=0)
    return(data)



# -*- coding: utf-8 -*-
#
# Simulation class
# Copyright (c) 2019-2022 Klaus K. Holst.  All rights reserved.

import kvaser as kv
import networkx as nx
import numpy as np
import pandas as pd
from PIL import Image # Python Imaging library
from io import BytesIO
import pydot # graphviz/dot
import patsy


class dag:
    r"""DAG model class

    Examples
    -----------

    >>> m = dag()
    >>> m.regression('y', ['x','z'])
    >>> m.regression('z', ['x'])
    >>> m.distribution('y', normal(scale=2))
    >>> m.distribution('z', poisson())
    >>> m.distribution('x', bernoulli())
    >>> print(m)
        y ~ x + z
        x ~ v
        z ~ x + w
        w ~ 1
        v ~ 1

        y: Normal distribution {'scale': 2}
        x: Binomial distribution {}
        z: Poisson distribution {}
        w: Discrete distribution {}
        v: Generic distribution {}

    >>> r = m.simulate(5000)
    >>> r
                      y    x    z    w    v
        0      7.482054  1.0  8.0  1.0  1.0
        1      1.324973  1.0  0.0 -1.0  1.0
        2      5.071795  1.0  4.0  1.0  1.0
        3     -2.643319  1.0  0.0 -1.0  1.0
        4      8.206240  1.0  4.0  1.0  1.0
        ...         ...  ...  ...  ...  ...
        4995   4.020084  1.0  3.0  0.0  2.0
        4996   0.277728  1.0  1.0 -1.0  2.0
        4997  12.546145  1.0  8.0  1.0  2.0
        4998   4.043192  1.0  0.0 -1.0  2.0
        4999   4.402478  1.0  1.0 -1.0  2.0

    >>> # import statsmodels.formula.api as smf
    >>> # import statsmodels.genmod.families as fam
    >>> # smf.glm('y ~ x+z', data=r, family=fam.Gaussian()).fit().summary()
    >>> # smf.glm('z ~ x', data=r, family=fam.Poisson()).fit().summary()
    >>> # smf.glm('x ~ 1', data=r, family=fam.Binomial()).fit().summary()
    """

    def __init__(self):
        r"""
        Constructor
        """
        self.G = nx.DiGraph()
        self._distribution = {}
        self._functionalform = {}

    def regression(self, y, x=[], f=None):
        r"""Add regression association between response 'y' and list of covariates 'x'

        Parameters
        ----------
        y: str
           response variable name
        x: list of str
           covariate names
        """
        self.G.add_node(y)
        if y not in self._distribution.keys():
            self.distribution(y)
        for v in x:
            self.G.add_edge(v, y)
            if v not in self._distribution.keys():
                self.distribution(v)
                self._functionalform[v] = f
        return self

    def distribution(self, y, generator=kv.normal()):
        r"""Set distribution of variable

        Parameters
        ----------
        y: str
           variable name
        generator: Dist inherited class
           random generator (see details below)

        Notes
        ----------
        Available generators:
          - kv.normal(scale=...)
          - kv.bernoulli()
          - kv.poisson()
          - kv.discrete(values=..., p=...)

        New generators can easily be constructed.

        >>> class mydist(kv.Dist):
        >>> @kv.randgen
        >>> def gen(self, param, **kwargs):
                return np.repeat([1,2], len(param)/2)

        'param' is a parameter array derived from the regression design -
        typically the mean parameter.

        Here a generator for a Gamma distribution:

        >>> class mydist2(kv.Dist):
        >>>    invlink = np.exp
        >>>    @kv.randgen
        >>>    def gen(self, param, rng, **kwargs):
        >>>        return rng.gamma(scale=param, shape=param, **kwargs)

        """
        self.G.add_node(y)
        self._distribution[y] = generator
        return self

    def simulate(self, n=1, p={}, file=None, rng=None):
        r"""Simulate from model

        Parameters
        ----------
        n: int
           number of samples to simulate
        p: dict
           (optional) Dictionary of parameters (see below)
        file: str
           (optional) file name to store data in CSV format
        rng: numpy.random._generator.Generator
           (optional) Random number generator

        Notes
        ----------
        Parameters are default 0 for all intercepts and 1 for all other regression coefficients.
        This can be changed for a subset of the parameters via the argument 'p'.
        'p' must be a dictionary where the keys are the names of the parameters where intercept parameters
        are simply the name of the variable, e.g., 'y', and the regression coefficients are named as
        'y~x' (response y and covariate x). For examples

        >>> m = kv.dag()
        >>> m.regression('y', ['x','z'])
        >>> m.distribution('y', kv.poisson())
        >>> p = {'y': -1, 'y~x': 2}
        >>> m.simulate(10, p=p)

        which draws 10 simulations from the model

        .. math:: y|x,z \sim \operatorname{pois}\left\{\exp(-1 + 2x + z)\right\}

        """
        deg = dict(self.G.in_degree)
        vv = list(self.G.nodes)
        n = int(n)
        res = np.ndarray((n, len(vv)))

        p0 = {}
        for y in vv:
            p0[y] = 0.0
            if y in p.keys():
                p0[y] = p[y]
            par = self.G.predecessors(y)
            for x in par:
                pname = y + '~' + x
                p0[pname] = 1
                if pname in p.keys():
                    p0[pname] = p[pname]

        while any(x>=0 for x in deg.values()):
            for v, d in deg.items():
                if d>=0:
                    par = list(self.G.predecessors(v))
                    subdict = dict((k, deg[k]) for k in par if k in deg)
                    if all(x<0 for x in subdict.values()):
                        deg[v] = -1
                        pos = vv.index(v)
                        lp = np.repeat([float(p0[v])], n)
                        for x in par:
                            pname = v + '~' + x
                            posx = int(vv.index(x))
                            lp += p0[pname]*res[:,posx]
                        y = np.float64(self._distribution[v].simulate(lp=lp, rng=rng))
                        res[:,pos] = y
        df = pd.DataFrame(res)
        df.columns = vv
        if file is not None:
            df.to_csv(file)
            return None
        return df

    def summary(self):
        r"""Summary method"""
        res = {}
        for v in self.G.nodes:
            parents = list(self.G.predecessors(v))
            if len(parents)==0:
                parents = "1"
            formula = v + ' ~ ' + ' + '.join(parents)
            dist = self._distribution[v]
            res[v] = (formula, dist)
        return res

    def design(self, data):
        r"""Extract model design matrix

        Parameters
        ----------
        data: pandas.DataFrame
              DataFrame with all variables in the model (as returned by the simulate method)

        Returns
        ----------
        Dict
           Dictonary which for each variable name contains a tuple
           y, X of response and design matrix values

        """
        res = {}
        mod = self.summary()
        for v in mod.keys():
            f = mod[v][0]
            y, X = patsy.dmatrices(f, data=data)
            res[v] = (y, X)
        return res

    def estimate(self, data):
        r"""Estimate model parameters

        Parameters
        ----------
        data: pandas.DataFrame
              DataFrame with all variables in the model (as returned by the simulate method)

        Returns
        ----------
        Dict
           Dictonary which for each variable name contains a dictonary with elements
           'coef' (regression coefficients) and 'ic' (influence curve).

        Examples
        ----------

        >>> m = kv.dag().regression('y', 'x').distribution('y', poisson())
        >>> d = m.simulate(1000)
        >>> m.estimate(d)['y']
            {'coef': array([-0.01545999,  0.96640726]), 'ic': ...}

        """
        res = {}
        mod = self.summary()
        for v in mod.keys():
            dist = mod[v][1]
            f = mod[v][0]
            y, X = patsy.dmatrices(f, data=data)
            coef, influ = dist.estimate(y=y.ravel(), X=X)
            res[v] = {'coef': coef, 'ic': influ}
        return res

    def plot(self):
        r"""Plot DAG model
        """
        pdot = nx.nx_pydot.to_pydot(self.G)
        Image.open(BytesIO(pdot.create_png())).show()

    def __str__(self):
        st = 'DAG model class'
        if len(self.G.nodes)>0:
            st += '\n'
        for f in self.summary().values():
            st += '\n' + f[0]
        for k,v in self._distribution.items():
            st += '\n' + str(k) + ': ' + str(v)
        return st





# encoding: utf-8
#
# Wrapper for interacting with Azure Blob Storage
# Copyright (c) 2020-2022 Klaus K. Holst.  All rights reserved.
#

import pandas as pd
import pickle
import os
from io import StringIO, BytesIO
from azure.storage.blob import BlobServiceClient
import azure.core as az
from .__utils__ import filesize # noqa F403
# import pyarrow.parquet as pq


class blob_storage:
    """Simple class for reading and writing files on Azure Blob Storage

    The connection string can either be given at the class
    initialization or alternatively it will be read from the
    environment variable AZURE_STORAGE_CONNECTION_STRING
    """
    blobservice = None
    active_container = None
    container_name = ""

    def __init__(self, container=None, connect_str=None,
                 account=None, credential=None, account_url=None, ):
        """Class initialization

        Examples
        --------
        import kvaser
        blob = kvaser.blob_storage('my-kvaser-test')
        df = kvaser.getdata() # A test pandas dataframe
        blob.write_pq(df, 'a.pq')
        {'etag': '"0x8D84D076069EDBB"', 'last_modified': datetime.datetime(2020, 8, 30, 17 ...
        blob.read_pq('a.pq')
        blob.write_csv(df, 'a.csv')
        blob.list()
        blob.read_csv('a.csv')

        Parameters
        ----------
        container: str
            Name of container to use
        connect_str: str
            Connection string (optional). If 'None' an attempt to retrieve this
            from the environment variable AZURE_STORAGE_CONNECTION_STRING will be made
        account: str
            Account (optional)
        credential: str
            SAS token (optional). If 'None' an attempt to retrieve this
            from the environment variable AZURE_SAS_TOKEN will be made
        account_url: str
            Account URL (optional)
        Returns
        -------
        blob_storage
            blob_storage object
        """
        if connect_str is None and credential is None:
            connect_str = os.getenv('AZURE_STORAGE_CONNECTION_STRING')
        if credential is None:
            credential = os.getenv('AZURE_SAS_TOKEN')

        if account_url is None:
            if account is None:
                account = os.getenv("OAUTH_STORAGE_ACCOUNT_NAME")
            account_url = "https://{}.blob.core.windows.net".format(account)

        if connect_str is not None:
            print("Connection string")
            self.blobservice = BlobServiceClient.from_connection_string(connect_str)
        else:
            # print("Credential")
            self.blobservice = BlobServiceClient(account_url=account_url,
                                                 credential=credential)

        if container is not None:
            self.select_container(container)

    def select_container(self, name):
        """Switch container.

        A new container will be created if the specified one does not yet exists.

        Parameters
        ----------
        name: str
            Name of container
        """
        self.container_name = name
        self.active_container = self.blobservice.get_container_client(name)
        try:
            self.active_container.get_container_properties()
        except az.exceptions.ResourceNotFoundError:
            self.active_container = self.active_container.create_container()

    def list(self):
        """List blobs in the container
        """
        for x in self.active_container.list_blobs():
            sz = filesize(x.size)
            print(x.name + '\t' + str(sz[0]) + ' ' + sz[1])

    def list_container(self):
        """List all containers
        """
        for x in self.blobservice.list_containers():
            print(x.name)

    def read_str(self, file):
        blob_client = self.blobservice.get_blob_client(container=self.container_name, blob=file)
        dl = blob_client.download_blob()
        return dl.content_as_text()

    def print(self, file):
        print(self.read_str(file))

    def write_str(self, string, file, append=True, sep=os.linesep):
        blob_client = self.blobservice.get_blob_client(container=self.container_name, blob=file)
        try:
            blob_client.get_blob_properties()
            if not append:
                blob_client.create_append_blob()
        except az.exceptions.ResourceNotFoundError:
            blob_client.create_append_blob()
        buf = StringIO()
        try:
            buf.writelines(str(string) + sep)
            val = blob_client.upload_blob(buf.getvalue(), blob_type='AppendBlob')
        finally:
            buf.close()
        return val

    def write(self, obj, file, type='pickle'):
        """Write pickle file on blob storage
        """
        blob_client = self.blobservice.get_blob_client(container=self.container_name, blob=file)
        buf = BytesIO()
        try:
            if type == 'parquet':
                obj.to_parquet(buf, index=False)
            else:
                pickle.dump(obj, buf)
            val = blob_client.upload_blob(buf.getvalue(), blob_type='BlockBlob', overwrite=True)
        finally:
            buf.close()
        return val

    def read(self, file, type='pickle'):
        """Read pickle file on blob storage
        """
        blob_client = self.blobservice.get_blob_client(container=self.container_name, blob=file)
        dl = blob_client.download_blob()
        obj = BytesIO(dl.content_as_bytes())
        if type == 'parquet':
            val = pd.read_parquet(obj)
        else:
            val = pickle.load(obj)
        return val

    def read_pq(self, file):
        """Read parquet file from blob storage
        """
        return self.read(file, type='parquet')

    def write_pq(self, df, file):
        """Write pandas dataframe to parquet file on blob storage
        """
        return self.write(df, file, type='parquet')

    def read_csv(self, file):
        """Read csv file from blob storage
        """
        blob_client = self.blobservice.get_blob_client(container=self.container_name, blob=file)
        dl = blob_client.download_blob()
        df = pd.read_csv(StringIO(dl.content_as_text()))
        return df

    def write_csv(self, df, file):
        """Write pandas dataframe to csv file on blob storage
        """
        blob_client = self.blobservice.get_blob_client(container=self.container_name, blob=file)
        output = df.to_csv(index=False, encoding='utf-8')
        return blob_client.upload_blob(output, blob_type='BlockBlob', overwrite=True)

    def delete(self, file):
        """Delete object (file) from blob storage
        """
        blob_client = self.blobservice.get_blob_client(container=self.container_name, blob=file)
        try:
            blob_client.delete_blob()
        except az.exceptions.ResourceNotFoundError:
            pass

    def delete_container(self, name=None):
        """Delete container
        """
        if name is None:
            name = self.container_name
        self.blobservice.delete_container(name)
